#include <WINDOWS.H>
#include <CSTDARG>
#include <CSTDIO>
#include <CMATH>
#include "UNITPALETTE.H"
#include "RESOURCE.H"

// ###################### STRUCTS ######################
struct FileList
{
	char* filename;
	FileList* next;
};

#pragma pack(1)

struct CellOffset;

struct BmpColor
{
	BYTE r;
	BYTE g;
	BYTE b;
	BYTE a;
};

struct BatchConvertToDC6Args
{
	size_t structsize;
	bool transparent;
	bool palette;
	int directions;
	char background;
	CellOffset* offsets;
};

struct BmpHeader
{
	WORD magic;
	DWORD size;
	WORD not_used1;
	WORD not_used2;
	DWORD data_offset;
	DWORD info_size;
	DWORD width;
	DWORD height;
	WORD planes;
	WORD bits_per_pixel;
	DWORD compression;
	DWORD image_size;
	DWORD pixels_per_meter_w;
	DWORD pixels_per_meter_h;
	DWORD palette_size;
	DWORD color_important;
};

struct PcxHeader
{
	BYTE manufacturer;
	BYTE version;
	BYTE encoding;
	BYTE bits_per_pixel;
	WORD left;
	WORD top;
	WORD width;
	WORD height;
	WORD vres;
	WORD hres;
	BYTE palette[48];
	BYTE reserved;
	BYTE color_planes;
	WORD bytes_per_line;
	WORD palette_type;
	BYTE filler[58];
};

struct BmpFile
{
	BmpHeader* head;
	BYTE* pixels;
	DWORD size;
	CellColor* palette;
};

struct PcxFile
{
	PcxHeader* head;
	BYTE* pixels;
	DWORD size;
	CellColor* palette;
};

struct CellHeader
{
	DWORD version;
	DWORD unknown1;
	DWORD unknown2;
	DWORD filler;
	DWORD dir_count;
	DWORD frame_count;
};

struct CellFrame
{
	DWORD flip;
	DWORD width;
	DWORD height;
	long x_offset;
	long y_offset;
	DWORD unknown;
	DWORD next_frame_offset;
	DWORD frame_size;
};

struct CellFile;

struct CellFrameList
{
	CellFrame* frame;
	BYTE* pixels;
	DWORD size;
	CellFrameList* next;
	DWORD offset;
	CellFile* parent;
};

struct CellOffset
{
	long x_offset;
	long y_offset;
};

struct CellFile
{
	CellHeader* head;
	CellFrameList* framehead;
	CellFrameList* frametail;
	DWORD frame_count;
	DWORD dir_count;
	DWORD width;
	DWORD height;
	bool transparent;
	char filler;
};

#pragma pack()

// ###################### TYPEDEFS ####################
typedef int (__fastcall* FileListCycler)(FileList*, void*);

// ##################### PROTOTYPES ####################
void cleanup();

// ###################### GLOBALS ######################
static void* heap = 0;
static FileList* filehead = 0;
static FileList* filetail = 0;
static const DWORD pcxheadsize = sizeof(PcxHeader);
static bool debug_palette = false;






void __cdecl error(char* format, ...)
{
	va_list args;
	va_start(args, format);
	char buffer[500];
	memset(buffer, 0, sizeof(buffer));
	_vsnprintf(buffer, sizeof(buffer) - 1, format, args);
	char message[1000];
	memset(message, 0, sizeof(message));
	_snprintf(message, sizeof(message) - 1, "Error: %s", buffer);
	MessageBox(0, message, "Fatal Error", MB_OK | MB_ICONERROR);
	cleanup();
	void* process = GetCurrentProcess();
	TerminateProcess(process, -1);
}

long __fastcall GetPaletteIndex(CellColor* palette, DWORD colors, BYTE red, BYTE green, BYTE blue, bool no_act_color, int filler)
{
	if (palette == 0 || colors == 0)
		return 0;

	long index = 0;
	DWORD diff = 0xFFFFFFFF;
	DWORD color = 0;

	while (color < colors)
	{
		if (no_act_color == true)
		{
			if (color >= 224 && color <= 254)
			{
				color = 255;
				continue;
			}
		}

		if ((int)color != filler)
		{
			DWORD dr = palette[color].r - red;
			DWORD dg = palette[color].g - green;
			DWORD db = palette[color].b - blue;

			DWORD this_diff = abs(((dr*dr)*76)+((dg*dg)*150)+((db*db)*28));

			if (this_diff <= diff)
			{
				index = color;
				diff = this_diff;
			}

		}

		color++;
	}	

	return index;
}

void InitHeap(int megs)
{
	int size = megs << 20;
	if (size <= 0)
		error("HEAP CREATION FAILURE, HEAPSIZE: %u", size);

	heap = HeapCreate(0, size, 0);
	if (heap == 0)
		error("HEAP CREATION FAILURE");
}

bool KillHeap()
{
	if (heap == 0)
		return false;

	return (HeapDestroy(heap) != 0);
}

void* __fastcall silentalloc(int size, int count = 0)
{
	if (heap == 0)
		return 0;

	if (count != 0)
		size *= count;

	if (size <= 0)
		return 0;

	return HeapAlloc(heap, HEAP_ZERO_MEMORY, size);
}

void* __fastcall alloc(int size, int count = 0)
{
	if (heap == 0)
		error("MEMORY ALLOCATION FAILURE, INVALID HEAP");

	if (count != 0)
		size *= count;

	if (size <= 0)
		error("MEMORY ALLOCATION FAILURE, MEMSIZE: %u", size);

	void* mem = HeapAlloc(heap, HEAP_ZERO_MEMORY, size);
	if (mem == 0)
		error("MEMORY ALLOCATION FAILURE");

	return mem;
}

void __fastcall clear(void* mem)
{
	if (mem == 0 || heap == 0)
		return;

	HeapFree(heap, 0, mem);
}

bool __fastcall getext(char* filename, int size, char** ext)
{
	if (filename == 0 || ext == 0 || size <= 0)
		return false;

	*ext = 0;
	for (int i = size; i >= 0; i--)
	{
		if (filename[i] == '.')
		{
			int offset = i + 1;
			if (offset >= size)
				return false;

			*ext = filename + offset;
			return true;
		}
	}

	return false;
}

bool __fastcall swapext(char* filename, char* ext, int size)
{
	if (filename == 0 || ext == 0 || size <= 0)
		return false;

	char* oldext = 0;
	int extlen = lstrlen(ext);
	if (getext(filename,size,&oldext) == 0)
		return false;

	int oldlen = lstrlen(oldext);
	if (oldlen < extlen)
		return false;

	*oldext = 0;
	lstrcat(filename,ext);
	return true;
}

void DeleteFileList()
{
	FileList* file = filehead;
	filetail = 0;
	filehead = 0;

	while (file != 0)
	{
		FileList* next = file->next;
		file->next = 0;

		if (file->filename != 0)
		{
			clear(file->filename);
			file->filename = 0;
		}

		clear(file);
		file = next;
	}
}

void CreateFileNode(char* path)
{
	if (path == 0)
		return;

	int length = lstrlen(path);
	if (length <= 0)
		return;

	char* buff = (char*)alloc(sizeof(char), length + 1);
	lstrcpy(buff, path);

	FileList* node = (FileList*)alloc(sizeof(FileList));
	if (filehead == 0)
		filehead = node;

	if (filetail != 0)
		filetail->next = node;

	node->filename = buff;
	filetail = node;
}

int CreateFileList(char* filter)
{
	WIN32_FIND_DATA fd;
	memset(&fd, 0, sizeof(fd));

	char buffer[MAX_PATH];
	memset(buffer, 0, sizeof(buffer));
	GetCurrentDirectory(sizeof(buffer) - 1, buffer);

	char search[MAX_PATH];
	_snprintf(search, sizeof(search) - 1, "%s\\*.%s", buffer, filter);

	int files = 0;
	void* find = FindFirstFile(search, &fd);
	if (find == 0)
		error("FAILED OPENING SEARCH HANDLE");

	do {

		char* filename = fd.cFileName;
		if (filename != 0)
		{
			int length = lstrlen(filename);
			if (length > 0)
			{
				char* ext = 0;
				if (getext(filename, length, &ext) != false)
				{
					if (lstrcmpi(ext, filter) == 0)
					{
						CreateFileNode(filename);
						files++;
					}
				}
			}
		}

	} while (FindNextFile(find, &fd) != 0);

	FindClose(find);
	return files;
}

size_t __fastcall TellSize(FILE* buff)
{
	if (buff == 0)
		return 0;

	fseek(buff, 0, SEEK_END);
	size_t size = ftell(buff);
	rewind(buff);

	return size;
}

bool ValidateBMPHead(BmpHeader* head)
{
	if (head == 0)
		return false;

	if (head->magic != 0x4D42)
		return false;

	if (head->bits_per_pixel != 8)
		return false;

	if (head->compression != 0)
		return false;

/*
	// BMP files made by XNView and other tools ignore these fields
	if (head->palette_size != 256 || head->color_important != 256)
		return false;
*/

	return true;
}

bool ValidateBMP(BmpFile* file)
{
	if (file == 0)
		return false;

	if (ValidateBMPHead(file->head) == false)
		return false;

	if (file->palette == 0 || file->pixels == 0 || file->size == 0)
		return false;

	if (file->size != file->head->width * file->head->height)
		return false;

	return true;
}

void DeleteBMP(BmpFile* file)
{
	if (file == 0)
		return;

	if (file->head != 0)
		clear(file->head);
	file->head = 0;

	if (file->palette != 0)
		clear(file->palette);
	file->palette = 0;

	if (file->pixels != 0)
		clear(file->pixels);
	file->pixels = 0;

	clear(file);
}

BmpFile* ReadBMP(char* filename, char filler)
{
	if (filename == 0)
		return 0;

	FILE* buff = fopen(filename, "rb");
	if (buff == 0)
		return 0;

	int size = TellSize(buff);
	if (size <= 0)
	{
		fclose(buff);
		return 0;
	}

	BmpColor* bmp_palette = (BmpColor*)silentalloc(sizeof(BmpColor), 256);
	if (bmp_palette == 0)
	{
		fclose(buff);
		return 0;
	}

	CellColor* palette = (CellColor*)silentalloc(sizeof(CellColor), 256);
	if (palette == 0)
	{
		clear(bmp_palette);
		fclose(buff);
		return 0;
	}

	BmpFile* bmpfile = (BmpFile*)silentalloc(sizeof(BmpFile));
	if (bmpfile == 0)
	{
		clear(bmp_palette);
		clear(palette);
		fclose(buff);
		return 0;
	}

	BmpHeader* bmphead = (BmpHeader*)silentalloc(sizeof(BmpHeader));
	if (bmphead == 0)
	{
		clear(bmp_palette);
		clear(palette);
		clear(bmpfile);
		fclose(buff);
		return 0;
	}

	int bytes_read = fread(bmphead, 1, sizeof(BmpHeader), buff);
	size -= bytes_read;
	if (bytes_read != sizeof(BmpHeader) || ValidateBMPHead(bmphead) == false || size <= 0)
	{
		clear(bmp_palette);
		clear(palette);
		clear(bmpfile);
		clear(bmphead);
		fclose(buff);
		return 0;
	}

	bytes_read = fread(bmp_palette, 1, 256 * sizeof(BmpColor), buff);
	size -= bytes_read;
	if (bytes_read != 256 * sizeof(BmpColor) || size <= 0)
	{
		clear(bmp_palette);
		clear(palette);
		clear(bmpfile);
		clear(bmphead);
		fclose(buff);
		return 0;
	}

	for (int i = 0; i < 256; i++)
	{
		palette[i].b = bmp_palette[i].r;
		palette[i].g = bmp_palette[i].g;
		palette[i].r = bmp_palette[i].b;
	}

	clear(bmp_palette);
	bmp_palette = 0;

	BYTE* pixels = (BYTE*)silentalloc(size);
	if (pixels == 0)
	{
		clear(palette);
		clear(bmpfile);
		clear(bmphead);
		fclose(buff);
		return 0;
	}

	memset(pixels, filler, size);
	bytes_read = fread(pixels, 1, size, buff);
	if (bytes_read != size || (DWORD)size != bmphead->width * bmphead->height)
	{
		clear(pixels);
		clear(palette);
		clear(bmpfile);
		clear(bmphead);
		fclose(buff);
		return 0;
	}

	bmpfile->head = bmphead;
	bmpfile->palette = palette;
	bmpfile->pixels = pixels;
	bmpfile->size = bytes_read;

	fclose(buff);
	return bmpfile;
}

PcxFile* ReadPCX(char* filename, char filler)
{
	if (filename == 0)
		return 0;

	FILE* buff = fopen(filename, "rb");
	if (buff == 0)
		return 0;

	int size = TellSize(buff);
	if (size <= 0)
	{
		fclose(buff);
		return 0;
	}

	PcxFile* pcxfile = (PcxFile*)silentalloc(sizeof(PcxFile));
	if (pcxfile == 0)
	{
		fclose(buff);
		return 0;
	}

	PcxHeader* pcxhead = (PcxHeader*)silentalloc(sizeof(PcxHeader));
	if (pcxhead == 0)
	{
		clear(pcxfile);
		fclose(buff);
		return 0;
	}

	int bytes_read = fread(pcxhead, 1, pcxheadsize, buff);
	if (bytes_read != pcxheadsize)
	{
		clear(pcxfile);
		clear(pcxhead);
		fclose(buff);
		return 0;
	}

	size -= bytes_read;
	if (size <= 0)
	{
		clear(pcxfile);
		clear(pcxhead);
		fclose(buff);
		return 0;
	}

	BYTE* pixels = (BYTE*)silentalloc(size);
	if (pixels == 0)
	{
		clear(pcxfile);
		clear(pcxhead);
		fclose(buff);
		return 0;
	}

	memset(pixels, filler, size);
	bytes_read = fread(pixels, 1, size, buff);
	if (bytes_read != size)
	{
		clear(pixels);
		clear(pcxfile);
		clear(pcxhead);
		fclose(buff);
		return 0;
	}

	pcxfile->head = pcxhead;
	pcxfile->pixels = pixels;
	pcxfile->size = bytes_read;
	fclose(buff);
	return pcxfile;
}

void DeletePCX(PcxFile* pcxfile)
{
	if (pcxfile == 0)
		return;
	if (pcxfile->head != 0)
		clear(pcxfile->head);
	pcxfile->head = 0;
	if (pcxfile->pixels != 0)
		clear(pcxfile->pixels);
	pcxfile->pixels = 0;
	if (pcxfile->palette != 0)
		clear(pcxfile->palette);
	pcxfile->palette = 0;
	clear(pcxfile);
}

bool ValidatePCX(PcxFile* pcxfile)
{
	if (pcxfile == 0)
		return false;

	if (pcxfile->head == 0)
		return false;

	if (pcxfile->pixels == 0 || pcxfile->size == 0)
		return false;

	if (pcxfile->head->manufacturer != 10)
		return false;

	if (pcxfile->head->version != 5)
		return false;

	if (pcxfile->head->encoding != 1)
		return false;

	if (pcxfile->head->bits_per_pixel != 8)
		return false;

	return true;
}

bool GetPCXPalette(PcxFile* pcxfile, int in_size, BYTE* data)
{
	if (pcxfile == 0 || in_size <= 0 || data == 0 || *data != 12)
		return false;

	if (pcxfile->palette != 0)
		return false;

	if (--in_size < 0)
		return false;
	data++;

	CellColor* palette = (CellColor*)silentalloc(sizeof(CellColor), 256);
	if (palette == 0)
		return false;

	for (int i = 0; i < 256; i++)
	{
		palette[i].r = data[0];
		palette[i].g = data[1];
		palette[i].b = data[2];

		in_size -= 3;
		if (in_size < 0)
		{
			clear(palette);
			return false;
		}

		data += 3;		
	}

	pcxfile->palette = palette;
	return true;
}

bool DecodeBMP(BmpFile* bmpfile)
{
	if (ValidateBMP(bmpfile) == false)
		return false;

	BYTE* new_pixels = (BYTE*)silentalloc(bmpfile->size);
	if (new_pixels == 0)
		return false;

	int w = bmpfile->head->width;
	int h = bmpfile->head->height - 1;

	int line = 0;
	for (int y = h; y >= 0; y--)
	{
		int offset = y * w;
		int offset_new = line * w;

		for (int x = 0; x < w; x++)
			new_pixels[offset_new+x] = bmpfile->pixels[offset+x];

		line++;
	}

	clear(bmpfile->pixels);
	bmpfile->pixels = new_pixels;
	return true;
}

bool DecodePCX(PcxFile* pcxfile)
{
	if (ValidatePCX(pcxfile) == false)
		return false;

	int size = (pcxfile->head->width + 1) * (pcxfile->head->height + 1);
	if (size <= 0)
		return false;

	BYTE* buff = (BYTE*)silentalloc(size);
	if (buff == 0)
		return false;

	BYTE* in		= pcxfile->pixels;
	BYTE* out		= buff;
	int out_size	= size;
	int in_size		= pcxfile->size;
	
	for (int y = 0; y <= pcxfile->head->height; y++)
	{
		int x = 0;
		while (x <= pcxfile->head->width)
		{
			BYTE pixel = *in;
			if (--in_size < 0)			// buffer overflow prevention
			{
				clear(buff);
				return false;
			}
			in++;

			int runlength = 1;
			if ((pixel & 192) == 192)
			{
				runlength = pixel & 63;

				pixel = *in;
				if (--in_size < 0)		// buffer overflow prevention
				{
					clear(buff);
					return false;
				}
				in++;
			}

			while (runlength-- > 0)
			{
				*out = pixel;
				if (--out_size < 0)		// buffer overflow prevention
				{
					clear(buff);
					return false;
				}
				out++;

				x++;
			}
		}
	}

	if (GetPCXPalette(pcxfile, in_size, in) == false)
	{
		clear(buff);
		return false;
	}

	clear(pcxfile->pixels);
	pcxfile->pixels = buff;
	pcxfile->size = size;
	pcxfile->head->width++;
	pcxfile->head->height++;
	return true;
}

void PixelToASCII(char* filename, BYTE* pixels, int size, int rows)
{
	if (filename == 0 || pixels == 0 || size <= 0)
		return;

	int len = lstrlen(filename);
	if (len <= 0)
		return;

	char* newname = (char*)alloc(len+1);
	if (newname == 0)
		return;

	lstrcpy(newname, filename);
	if (swapext(newname, "txt", len) == false)
	{
		clear(newname);
		return;
	}

	FILE* outfile = fopen(newname, "w");
	if (outfile == 0)
	{
		clear(newname);
		return;
	}

	int row = 0;
	for (int i = 0; i < size; i++)
	{
		char pixel = pixels[i];
		if (pixel != 0) {
			pixel = '#';
		} else {
			pixel = '.';
		}
		fprintf(outfile,"%c",pixel);
		row++;
		if (row >= rows)
		{
			row = 0;
			fprintf(outfile,"\n");
		}
	}

	fclose(outfile);
	clear(newname);
}

long __fastcall GetDC6FrameCount(long width, long height)
{
	double dwidth = (double)width;
	double dheight = (double)height;
	double dframes = ceil(dwidth / 256) * ceil(dheight / 256);
	return (long)dframes;
}

void DeleteDC6(CellFile* cellfile)
{
	if (cellfile == 0)
		return;

	if (cellfile->head != 0)
		clear(cellfile->head);
	cellfile->head = 0;

	CellFrameList* frame = cellfile->framehead;
	cellfile->framehead = 0;
	cellfile->frametail = 0;
	while (frame != 0)
	{
		CellFrameList* next = frame->next;
		frame->next = 0;
		if (frame->frame != 0)
			clear(frame->frame);
		frame->frame = 0;
		frame->parent = 0;
		clear(frame);
		frame = next;
	}

	clear(cellfile);
}

bool AddDC6Frame(CellFile* file, long width, long height, BYTE* pixels)
{
	if (file == 0 || width <= 0 || height <= 0 || pixels == 0)
		return false;

	CellFrameList* cellframe = (CellFrameList*)silentalloc(sizeof(CellFrameList));
	if (cellframe == 0)
		return false;

	if (file->framehead == 0)
		file->framehead = cellframe;

	if (file->frametail != 0)
		file->frametail->next = cellframe;

	cellframe->parent = file;
	file->frametail = cellframe;
	long frame_size = width * height;
	if (frame_size <= 0)
		return 0;

	cellframe->pixels = pixels;
	cellframe->size = frame_size;

	CellFrame* framehead = (CellFrame*)silentalloc(sizeof(CellFrame));
	if (framehead == 0)
		return false;

	framehead->flip = 0;
	framehead->width = width;
	framehead->height = height;
	framehead->unknown = 0;
	framehead->x_offset = 0;
	framehead->y_offset = 0;
	framehead->next_frame_offset = 0;
	framehead->frame_size = 0;
	cellframe->frame = framehead;
	file->frame_count++;
	return true;
}

CellFile* CreateDC6File(BYTE* pixels, DWORD size, long width, long height, bool transparent, char fillcolor, int directions, CellOffset* offsets)
{
	if (pixels == 0 || size == 0)
		return 0;

	if (width <= 0 || height <= 0 || directions < 1 || directions > 64)
		return 0;
		
	CellFile* cellfile = (CellFile*)silentalloc(sizeof(CellFile));
	if (cellfile == 0)
		return 0;

	CellHeader* cellhead = (CellHeader*)silentalloc(sizeof(CellHeader));
	cellfile->head = cellhead;
	cellfile->height = height;
	cellfile->width = width;
	cellfile->transparent = transparent;
	cellfile->filler = fillcolor;
	cellfile->dir_count = directions;

	if (cellhead == 0)
	{
		DeleteDC6(cellfile);
		return 0;
	}

	long frames = GetDC6FrameCount(width, height);
	if (frames <= 0 || frames % directions != 0)
	{
		DeleteDC6(cellfile);
		return 0;
	}

	cellhead->version = 6;
	cellhead->unknown1 = 1;				// cell flags
	cellhead->unknown2 = 0;				// encoding type
	cellhead->filler = 0xEEEEEEEE;
	cellhead->dir_count = directions;
	cellhead->frame_count = frames / directions;

	for (int y = 0; y < height; y += 256)
	{
		int x = 0;
		while (x < width)
		{
			long frame_x = (width - x > 256) ? 256 : width - x;
			long frame_y = (height - y > 256) ? 256 : height - y;
			DWORD offset = x + (y * width);

			BYTE* frame_px = pixels + offset;
			long frame_size = size - offset;
			if (frame_size <= 0)
			{
				DeleteDC6(cellfile);
				return 0;
			}

			if (AddDC6Frame(cellfile, frame_x, frame_y, frame_px) == false)
			{
				DeleteDC6(cellfile);
				return 0;
			}

			x += 256;
		}
	}

	if (offsets != 0)
	{
		DWORD this_dir = 0;
		DWORD this_frame = 0;
		CellFrameList* frame = cellfile->framehead;
		while (frame != 0)
		{
			frame->frame->x_offset = offsets[this_dir].x_offset;
			frame->frame->y_offset = offsets[this_dir].y_offset;
			frame = frame->next;
			if (frame == 0)
				break;

			this_frame++;
			if (this_frame >= cellfile->head->frame_count)
			{
				this_dir++;
				if (this_dir >= cellfile->head->dir_count)
				{
					DeleteDC6(cellfile);
					return 0;
				}

				this_frame = 0;
			}
		}
	}

	return cellfile;
}

bool ValidateDC6(CellFile* file)
{
	if (file == 0)
		return false;

	if (file->width == 0 || file->height == 0)
		return false;

	if (file->frame_count <= 0)
		return false;

	if (file->framehead == 0)
		return false;

	if (file->frametail == 0)
		return false;

	CellHeader* head = file->head;
	if (head == 0)
		return false;

	if (head->dir_count <= 0 || head->frame_count <= 0)
		return false;

	if (head->version != 6)
		return false;

	CellFrameList* frame = file->framehead;
	DWORD frames = 0;
	while (frame != 0)
	{
		if (frame->parent == 0)
			return false;

		if (frame->pixels == 0)
			return false;

		if (frame->size <= 0)
			return false;

		CellFrame* framehead = frame->frame;
		if (framehead == 0)
			return false;

		if (framehead->width <= 0 || framehead->height <= 0)
			return false;

		frame = frame->next;
		frames++;
	}

	if (frames != head->frame_count * head->dir_count)
		return false;

	if (frames != file->frame_count)
		return false;

	return true;
}

bool write_data(FILE* file, size_t size, DWORD data)
{
	if (file != 0)
	{
		void* data_ptr = &data;
		DWORD written = fwrite(data_ptr, 1, size, file);
		if (written != size)
		{
			fclose(file);
			return false;
		}

		return true;
	}

	return false;
}

bool WriteDC6Head(CellHeader* head, FILE* file)
{
	if (file == 0)
		return false;

	if (head == 0 )
	{
		fclose(file);
		return false;
	}

	DWORD written = fwrite(head, 1, sizeof(CellHeader), file);
	if (written != sizeof(CellHeader))
	{
		fclose(file);
		return false;
	}

	return true;
}

bool WriteDC6Pixels(FILE* file, BYTE* data, long width, long height, long rowbytes, bool transparent, char filler)
{
	if (file == 0)
		return false;

	if (data == 0 || width <= 0 || height <= 0)
	{
		fclose(file);
		return false;
	}

	long line = height - 1;
	long pixels = 0;

	BYTE* in = data + (rowbytes * line);
	while (line >= 0)
	{
		while (pixels < width)
		{
			long pwidth = width - pixels;
			if (pwidth > 127)
				pwidth = 127;

			if (transparent == true)
			{
				if (*in == filler)
				{
					int count = 0;
					BYTE* test = in;

					for (int i = pixels; i < width; i++)
					{
						if (*test != filler)
							break;

						count++;
						test++;
					}

					if ((count + pixels) == width)
					{
						pixels += count;
						in += count;
						continue;
					}

					if (count > 127)
						count = 127;

					pixels += count;
					in += count;
					fputc(count + 128, file);
					continue;
				}
					else
				{
					for (int i = 0; i < pwidth; i++)
					{
						if (in[i] == filler)
						{
							pwidth = i;
							break;
						}
					}
				}
			}

			fputc(pwidth, file);
			for (int i = 0; i < pwidth; i++)
			{
				fputc(*in++, file);
				pixels++;
			}
		}

		pixels = 0;
		line--;
		fputc(128, file);
		in = data + (rowbytes * line);
	}


	return true;
}

bool WriteDC6Frame(CellFrameList* frame, FILE* file, DWORD ptr_offset)
{
	if (file == 0)
		return false;

	if (frame == 0 || ptr_offset == 0)
	{
		fclose(file);
		return false;
	}

	CellFrame* head = frame->frame;
	if (head == 0)
	{
		fclose(file);
		return false;
	}

	DWORD frame_pos = ftell(file);
	fseek(file, ptr_offset, SEEK_SET);
	if (write_data(file, 4, frame_pos) == false)
		return false;

	fseek(file, frame_pos, SEEK_SET);

	DWORD written = fwrite(head, 1, sizeof(CellFrame), file);
	if (written != sizeof(CellFrame))
	{
		fclose(file);
		return false;
	}

	DWORD last_pos = ftell(file) - 8;
	
	CellFile* cellfile = frame->parent;
	if (cellfile == 0)
	{
		fclose(file);
		return false;
	}

	if (WriteDC6Pixels(file, frame->pixels, head->width, head->height, cellfile->width, cellfile->transparent, cellfile->filler) == false)
		return false;

	frame_pos = ftell(file);
	fseek(file, last_pos, SEEK_SET);

	if (write_data(file, 4, frame_pos + 3) == false)
		return false; 
	
	if (write_data(file, 4, frame_pos - last_pos - 8) == false)
		return false;

	fseek(file, frame_pos, SEEK_SET);
	for (int i = 0; i < 3; i++)
		fputc(0xEE, file);

	return true;
}

char* GetFirstEmptyFileName(char* name, char* newext)
{
	if (name == 0)
		return 0;

	int len = lstrlen(name);
	if (len <= 0)
		return 0;

	char* newname = (char*)silentalloc(len + 1);
	if (newname == 0)
		return 0;
	lstrcpy(newname, name);

	if (newext != 0)
	{
		if (swapext(newname, newext, len) == false)
		{
			clear(newname);
			return 0;
		}
	}

	FILE* file = fopen(newname, "rb");
	if (file == 0)
		return newname;
	fclose(file);
	file = 0;

	char* ext = 0;
	if (getext(newname, len, &ext) == false)
	{
		clear(newname);
		return 0;
	}

	int extlen = lstrlen(ext);
	if (extlen <= 0)
	{
		clear(newname);
		return 0;
	}
	
	char* extbuf = (char*)silentalloc(extlen + 1);
	if (extbuf == 0)
	{
		clear(newname);
		return 0;
	}

	lstrcpy(extbuf, ext);
	
	ext--;
	if ((DWORD)ext <= (DWORD)newname)
	{
		clear(newname);
		return 0;
	}
	
	*ext = 0;

	for (int i = 0; i <= 1000; i++)
	{
		char buffer[MAX_PATH];
		memset(buffer, 0, sizeof(buffer));
		_snprintf(buffer, sizeof(buffer) - 1, "%s_%03u.%s", newname, i, extbuf);

		file = fopen(buffer, "rb");

		if (file == 0)
		{
			clear(extbuf);
			clear(newname);
			len = lstrlen(buffer);
			newname = (char*)silentalloc(len + 1);
			if (newname == 0)
				return 0;
			
			lstrcpy(newname, buffer);
			return newname;
		}
			else
		{
			fclose(file);
			file = 0;
		}
	}

	clear(extbuf);
	clear(newname);
	return 0;
}

bool WriteDC6File(CellFile* file, char* filename)
{
	if (filename == 0)
		return false;

	if (ValidateDC6(file) == false)
		return false;

	char* newname = GetFirstEmptyFileName(filename, "dc6");
	if (newname == 0)
		return false;

	FILE* outfile = fopen(newname, "wb");
	clear(newname);
	if (outfile == 0)
		return false;

	if (WriteDC6Head(file->head, outfile) == false)
		return false;

	DWORD refoffset = ftell(outfile);
	for (int i = 0; i < (int)file->frame_count; i++)
	{
		if (write_data(outfile, 4, 0) == false)
			return false;
	}

	DWORD frame_idx = 0;
	CellFrameList* frame = file->framehead;
	while (frame != 0)
	{
		if (WriteDC6Frame(frame, outfile, frame_idx * 4 + refoffset) == false)
			return false;

		frame_idx++;
		frame = frame->next;
	}

	fclose(outfile);
	return true;
}

bool ConvertColor(BYTE* pixels, int size, CellColor* img_palette, CellColor* palette, long filler, bool transparent)
{ 
	if (img_palette == 0 || palette == 0 || pixels == 0 || size <= 0)
		return false;

	if (transparent == false)
		filler = -1;

	BYTE* shiftidx = (BYTE*)silentalloc(256);
	if (shiftidx == 0)
		return false;

	FILE* outfile = 0;
	if (debug_palette)
	{
		outfile = fopen("palette_debug.txt", "w");
		if (outfile == 0)
			return false;

		fprintf(outfile, "COLOR DIFFERENCES\n");
		for (int i = 0; i < 256; i++)
		{
			BYTE r = img_palette[i].r - palette[i].r;
			BYTE g = img_palette[i].g - palette[i].g;
			BYTE b = img_palette[i].b - palette[i].b;
			fprintf(outfile, "COLOR: %3i (%3u %3u %3u)\n", i, r, g, b);
		}

		fprintf(outfile, "\nPALETTE COLOR CONVERSION TABLE\n");
	}

	for (int i = 0; i < 256; i++)
	{
		BYTE r = img_palette[i].r;
		BYTE g = img_palette[i].g;
		BYTE b = img_palette[i].b;

		if (debug_palette)
			fprintf(outfile, "%3u %3u %3u -> ", r, g, b);

		int idx = i;

		if (i != filler)
		{
			idx = GetPaletteIndex(palette, 256, r, g, b, true, filler);
			if (idx < 0 || idx >= 256)
			{
				clear(shiftidx);
				return false;
			}
		}

		shiftidx[i] = idx;

		img_palette[i].r = palette[i].r;
		img_palette[i].g = palette[i].g;
		img_palette[i].b = palette[i].b;

		if (debug_palette)
			fprintf(outfile, "%3u %3u %3u\n", palette[idx].r, palette[idx].g, palette[idx].b);
	}

	if (debug_palette)
	{
		fclose(outfile);
		outfile = 0;
	}

	for (long p = 0; p < size; p++)
	{
		int idx = pixels[p];
		if (idx < 0 || idx >= 256)
		{
			clear(shiftidx);
			return false;
		}
		pixels[p] = shiftidx[idx];
	}

	clear(shiftidx);
	return true;
}

int __fastcall PCX2DC6(FileList* file, void* arg)
{
	if (file == 0 || file->filename == 0 || arg == 0)
		return 0;

	BatchConvertToDC6Args* args = (BatchConvertToDC6Args*)arg;
	if (args->structsize != sizeof(BatchConvertToDC6Args))
		return 0;

	PcxFile* pcxfile = ReadPCX(file->filename, args->background);
	if (pcxfile == 0)
		return 0;

	int ret = (DecodePCX(pcxfile) != false) ? 1 : 0;
	if (ret == 1)
	{
		if (args->palette != false && ConvertColor(pcxfile->pixels, pcxfile->size, pcxfile->palette, UnitPalette, args->background, args->transparent) == false)
		{
			DeletePCX(pcxfile);
			return 0;
		}

		CellFile* cellfile = CreateDC6File(pcxfile->pixels, pcxfile->size, pcxfile->head->width, pcxfile->head->height, args->transparent, args->background, args->directions, args->offsets);
		if (cellfile == 0) {
			ret = 0;
		} else {
			WriteDC6File(cellfile, file->filename);
			DeleteDC6(cellfile);
		}
	}

	DeletePCX(pcxfile);
	return ret;
}

int __fastcall BMP2DC6(FileList* file, void* arg)
{
	if (file == 0 || file->filename == 0 || arg == 0)
		return 0;

	BatchConvertToDC6Args* args = (BatchConvertToDC6Args*)arg;
	if (args->structsize != sizeof(BatchConvertToDC6Args))
		return 0;

	BmpFile* bmpfile = ReadBMP(file->filename, args->background);
	if (bmpfile == 0)
		return 0;

	if (DecodeBMP(bmpfile) == false)
	{
		DeleteBMP(bmpfile);
		return 0;
	}

	if (args->palette != false && ConvertColor(bmpfile->pixels, bmpfile->size, bmpfile->palette, UnitPalette, args->background, args->transparent) == false)
	{
		DeleteBMP(bmpfile);
		return 0;
	}

	CellFile* cellfile = CreateDC6File(bmpfile->pixels, bmpfile->size, bmpfile->head->width, bmpfile->head->height, args->transparent, args->background, args->directions, args->offsets);
	if (cellfile != 0)
	{
		WriteDC6File(cellfile, file->filename);
		DeleteDC6(cellfile);
	}

	DeleteBMP(bmpfile);
	return 1;
}

int ProcessFileList(FileListCycler fpListFunc, void* args)
{
	if (IsBadCodePtr((FARPROC)fpListFunc) != 0)
		return 0;

	int count = 0;
	FileList* file = filehead;

	while (file != 0)
	{
		FileList* next = file->next;
		count += fpListFunc(file, args);
		file = next;
	}

	return count;
}

char* GetINIFilePath(char* filename)
{
	if (filename == 0 || lstrlen(filename) <= 0)
		return 0;

	char* buffer = (char*)alloc(MAX_PATH + 1);
	GetCurrentDirectory(MAX_PATH, buffer);

	int len = lstrlen(buffer);
	if (len <= 0)
	{
		clear(buffer);
		return 0;
	}

	char* path_buffer = (char*)silentalloc(MAX_PATH + 1);
	if (path_buffer == 0)
	{
		clear(buffer);
		return 0;
	}

	_snprintf(path_buffer, MAX_PATH, "%s\\%s", buffer, filename);
	clear(buffer);

	FILE* file = fopen(path_buffer, "r");
	if (file == 0)
	{
		clear(path_buffer);
		return 0;
	}

	fclose(file);
	return path_buffer;
}

char* GetMode(char* inifile)
{
	if (inifile == 0)
		return 0;

	char buffer[500];
	memset(buffer, 0, sizeof(buffer));

	if (GetPrivateProfileString("settings", "mode", "none", buffer, sizeof(buffer) - 1, inifile) == 0)
		return 0;

	int len = lstrlen(buffer);
	if (len <= 0)
		return 0;

	char* modebuff = (char*)silentalloc(len + 1);
	if (modebuff == 0)
		return 0;

	lstrcpy(modebuff, buffer);
	return modebuff;
}

bool __fastcall CheckUseConfig(char* szCmdLine)
{
	if (szCmdLine == 0)
		return false;

	int length = lstrlen(szCmdLine);
	if (length <= 0)
		return false;

	char* buffer = strstr(szCmdLine,"-usecfg");
	if (buffer == 0)
		return false;

	return true;
}

void LoadGlobalPalette(char* name)
{
	if (name == 0 || lstrlen(name) <= 0)
		return;

	char buffer[MAX_PATH];
	memset(buffer, 0, sizeof(buffer));
	if (GetCurrentDirectory(sizeof(buffer) - 1, buffer) == 0)
		return;

	char path[MAX_PATH];
	memset(path, 0, sizeof(path));
	_snprintf(path, sizeof(path) - 1, "%s\\Palettes\\%s", buffer, name);

	FILE* file = fopen(path, "rb");
	if (file == 0)
		return;

	if (TellSize(file) != sizeof(CellColor) * 256)
	{
		fclose(file);
		return;
	}

	for (int i = 0; i < 256; i++)
	{
		UnitPalette[i].b = fgetc(file);
		UnitPalette[i].g = fgetc(file);
		UnitPalette[i].r = fgetc(file);
	}

	fclose(file);
}

void UseConfigFile(char* inipath)
{
	char buffer[MAX_PATH];
	memset(buffer, 0, sizeof(buffer));

	BatchConvertToDC6Args args;
	args.structsize = sizeof(BatchConvertToDC6Args);

	args.transparent = (GetPrivateProfileInt("settings", "transparent", 0, inipath) != 0);
	args.directions = GetPrivateProfileInt("settings", "directions", 0, inipath);
	args.background = GetPrivateProfileInt("settings", "background", 0, inipath);
	args.palette = (GetPrivateProfileInt("settings", "palette", 0, inipath) != 0);
	args.offsets = 0;

	if (GetPrivateProfileInt("settings", "palette_debugfile", 0, inipath) != 0)
		debug_palette = true;

	if (args.palette != 0)
	{
		char name_buffer[MAX_PATH];
		memset(name_buffer, 0, sizeof(name_buffer));
		GetPrivateProfileString("settings", "palette_name", "", name_buffer, MAX_PATH - 1, inipath); 
		LoadGlobalPalette(name_buffer);
	}

	if (GetPrivateProfileInt("settings", "offsets", 0, inipath) != 0)
	{
		int directions = args.directions;
		if (directions <= 0)
			return;

		CellOffset* offsets = (CellOffset*)silentalloc(sizeof(CellOffset), directions);
		if (offsets == 0)
			return;	
						
		for (int i = 0; i < directions; i++)
		{
			int dir = i + 1;

			char offset_x[20];
			memset(offset_x, 0, sizeof(offset_x));
			_snprintf(offset_x, sizeof(offset_x) - 1, "x%u", dir);

			char offset_y[20];
			memset(offset_y, 0, sizeof(offset_y));
			_snprintf(offset_y, sizeof(offset_y) - 1, "y%u", dir);

			offsets[i].x_offset = GetPrivateProfileInt("settings", offset_x, 0, inipath);
			offsets[i].y_offset = GetPrivateProfileInt("settings", offset_y, 0, inipath);
		}

		args.offsets = offsets;
	}

	GetPrivateProfileString("settings", "filename", "", buffer, sizeof(buffer) - 1, inipath);
	int length = lstrlen(buffer);
	if (length <= 0)
	{
		CreateFileList("pcx");
		ProcessFileList(PCX2DC6, &args);
		DeleteFileList();
		CreateFileList("bmp");
		ProcessFileList(BMP2DC6, &args);
	}
		else
	{
		char* ext = 0;
		if (getext(buffer, length, &ext) == false)
			return;

		FILE* filetest = fopen(buffer, "rb");
		if (filetest == 0)
			return;

		fclose(filetest);
		CreateFileNode(buffer);

		if (lstrcmpi(ext, "pcx") == 0) {
			ProcessFileList(PCX2DC6, &args);
		} else if (lstrcmpi(ext, "bmp") == 0) {
			ProcessFileList(BMP2DC6, &args);
		}
	}

	if (args.offsets != 0)
		clear(args.offsets);
}

int CellConvProc(HWND hwnd, UINT msg, UINT wparam, LONG lparam)
{
	switch (msg)
	{
		case WM_INITDIALOG:
		{
			long icon = (long)LoadIcon(GetModuleHandle(0), MAKEINTRESOURCE(IDI_ICON1));
			SendMessage(hwnd,WM_SETICON,ICON_SMALL,icon);
			SendMessage(hwnd,WM_SETICON,ICON_BIG,icon);
			return 1;
		}

		case WM_CLOSE:
		{
			EndDialog(hwnd,0);
			return 1;
		}

		case WM_COMMAND:
		{
			switch (LOWORD(wparam))
			{
				case ID_BUTTON_GO:
				{
					CreateFileList("pcx");

					BatchConvertToDC6Args args;
					args.structsize = sizeof(BatchConvertToDC6Args);
					args.directions = 1;
					args.background = 0;
					args.transparent = 1;
					args.offsets = 0;
					args.palette = 0;
					int pcx_files = ProcessFileList(PCX2DC6, &args);

					DeleteFileList();
					CreateFileList("bmp");
					int bmp_files = ProcessFileList(BMP2DC6, &args);
					
					char buffer[500];
					memset(buffer, 0, sizeof(buffer));

					if (pcx_files + bmp_files > 0) {
						_snprintf(buffer, sizeof(buffer) -1, "Converted %d *.PCX file(s) and %d *.BMP file(s) to *.DC6!", pcx_files, bmp_files);
					} else {
						_snprintf(buffer, sizeof(buffer) - 1, "Couldn't find any *.PCX or *.BMP files to convert!");
					}


					MessageBox(hwnd, buffer, "CellConv (v1.1)", MB_OK);
					return 1;
				}

				case ID_BUTTON_QUIT:
				{
					EndDialog(hwnd,0);
					return 1;
				}
			}

			return 0;
		}
	}

	return 0;
}

void UseDialogBox(HINSTANCE hInstance)
{
	DialogBox(hInstance,MAKEINTRESOURCE(IDD_DIALOG_CELLCONV),0,CellConvProc);
}

int WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, char* szCmdLine, int nShowCmd)
{
	InitHeap(16);

	if (CheckUseConfig(szCmdLine) != false)
	{
		char* inipath = GetINIFilePath("CellConv.ini");
		if (inipath == 0)
			error("COULDN'T FIND CELLCONV.INI");

		UseConfigFile(inipath);
		clear(inipath);
	}
		else
	{
		UseDialogBox(hInstance);
	}

	cleanup();
	return 0;
}

void cleanup()
{
	DeleteFileList();
	KillHeap();
}